@@ -1,4 +1,4 @@
-use crate::client::RedisClient;
+use crate::clients::RedisClient;
 use crate::error::*;
 use crate::modules::backchannel::Backchannel;
 use crate::multiplexer::SentCommand;
@@ -6,6 +6,7 @@ use crate::protocol::types::DefaultResolver;
 use crate::protocol::types::RedisCommand;
 use crate::types::*;
 use crate::utils;
+use arc_swap::ArcSwap;
 use parking_lot::RwLock;
 use std::collections::VecDeque;
 use std::sync::atomic::AtomicUsize;
@@ -58,9 +59,116 @@ impl MultiPolicy {
   }
 }
 
+/// A lock-free internal representation of the performance config options from the `RedisConfig`.
+#[derive(Debug)]
+pub struct InternalPerfConfig {
+  pipeline: ArcSwap<bool>,
+  max_command_attempts: Arc<AtomicUsize>,
+  default_command_timeout: Arc<AtomicUsize>,
+  max_feed_count: Arc<AtomicUsize>,
+  cluster_cache_update_delay_ms: Arc<AtomicUsize>,
+  disable_auto_backpressure: ArcSwap<bool>,
+  disable_backpressure_scaling: ArcSwap<bool>,
+  min_sleep_duration: Arc<AtomicUsize>,
+  max_in_flight_commands: Arc<AtomicUsize>,
+}
+
+impl<'a> From<&'a RedisConfig> for InternalPerfConfig {
+  fn from(config: &'a RedisConfig) -> Self {
+    InternalPerfConfig {
+      pipeline: ArcSwap::from(Arc::new(config.performance.pipeline)),
+      max_command_attempts: Arc::new(AtomicUsize::new(config.performance.max_command_attempts as usize)),
+      default_command_timeout: Arc::new(AtomicUsize::new(config.performance.default_command_timeout_ms as usize)),
+      max_feed_count: Arc::new(AtomicUsize::new(config.performance.max_feed_count as usize)),
+      cluster_cache_update_delay_ms: Arc::new(AtomicUsize::new(
+        config.performance.cluster_cache_update_delay_ms as usize,
+      )),
+      disable_auto_backpressure: ArcSwap::from(Arc::new(config.performance.backpressure.disable_auto_backpressure)),
+      disable_backpressure_scaling: ArcSwap::from(Arc::new(
+        config.performance.backpressure.disable_backpressure_scaling,
+      )),
+      min_sleep_duration: Arc::new(AtomicUsize::new(
+        config.performance.backpressure.min_sleep_duration_ms as usize,
+      )),
+      max_in_flight_commands: Arc::new(AtomicUsize::new(
+        config.performance.backpressure.max_in_flight_commands as usize,
+      )),
+    }
+  }
+}
+
+impl InternalPerfConfig {
+  pub fn pipeline(&self) -> bool {
+    *self.pipeline.load().as_ref()
+  }
+
+  pub fn max_command_attempts(&self) -> usize {
+    utils::read_atomic(&self.max_command_attempts)
+  }
+
+  pub fn default_command_timeout(&self) -> usize {
+    utils::read_atomic(&self.default_command_timeout)
+  }
+
+  pub fn max_feed_count(&self) -> usize {
+    utils::read_atomic(&self.max_feed_count)
+  }
+
+  pub fn cluster_cache_update_delay_ms(&self) -> usize {
+    utils::read_atomic(&self.cluster_cache_update_delay_ms)
+  }
+
+  pub fn disable_auto_backpressure(&self) -> bool {
+    *self.disable_auto_backpressure.load().as_ref()
+  }
+
+  pub fn disable_backpressure_scaling(&self) -> bool {
+    *self.disable_backpressure_scaling.load().as_ref()
+  }
+
+  pub fn min_sleep_duration(&self) -> usize {
+    utils::read_atomic(&self.min_sleep_duration)
+  }
+
+  pub fn max_in_flight_commands(&self) -> usize {
+    utils::read_atomic(&self.max_in_flight_commands)
+  }
+
+  pub fn update(&self, config: &PerformanceConfig) {
+    self.pipeline.store(Arc::new(config.pipeline));
+    self
+      .disable_backpressure_scaling
+      .store(Arc::new(config.backpressure.disable_backpressure_scaling));
+    self
+      .disable_auto_backpressure
+      .store(Arc::new(config.backpressure.disable_auto_backpressure));
+
+    utils::set_atomic(&self.max_command_attempts, config.max_command_attempts as usize);
+    utils::set_atomic(
+      &self.default_command_timeout,
+      config.default_command_timeout_ms as usize,
+    );
+    utils::set_atomic(&self.max_feed_count, config.max_feed_count as usize);
+    utils::set_atomic(
+      &self.cluster_cache_update_delay_ms,
+      config.cluster_cache_update_delay_ms as usize,
+    );
+    utils::set_atomic(
+      &self.min_sleep_duration,
+      config.backpressure.min_sleep_duration_ms as usize,
+    );
+    utils::set_atomic(
+      &self.max_in_flight_commands,
+      config.backpressure.max_in_flight_commands as usize,
+    );
+  }
+}
+
 pub struct RedisClientInner {
   /// The client ID as seen by the server.
   pub id: Arc<String>,
+  /// The RESP version used by the underlying connections.
+  pub resp_version: Arc<ArcSwap<RespVersion>>,
   /// The response policy to apply when the client is in a MULTI block.
   pub multi_block: RwLock<Option<MultiPolicy>>,
   /// The state of the underlying connection.
@@ -81,6 +189,8 @@ pub struct RedisClientInner {
   pub keyspace_tx: RwLock<VecDeque<UnboundedSender<KeyspaceEvent>>>,
   /// An mpsc sender for reconnection events to `on_reconnect` streams.
   pub reconnect_tx: RwLock<VecDeque<UnboundedSender<RedisClient>>>,
+  /// An mpsc sender for cluster change notifications.
+  pub cluster_change_tx: RwLock<VecDeque<UnboundedSender<Vec<ClusterStateChange>>>>,
   /// MPSC senders for `on_connect` futures.
   pub connect_tx: RwLock<VecDeque<OneshotSender<Result<(), RedisError>>>>,
   /// A join handle for the task that sleeps waiting to reconnect.
@@ -99,6 +209,8 @@ pub struct RedisClientInner {
   pub backchannel: Arc<AsyncRwLock<Backchannel>>,
   /// The server host/port resolved from the sentinel nodes, if known.
   pub sentinel_primary: RwLock<Option<Arc<String>>>,
+  /// The internal representation of the performance config options from the `RedisConfig`.
+  pub perf_config: Arc<InternalPerfConfig>,
 
   /// Command latency metrics.
   #[cfg(feature = "metrics")]
@@ -120,6 +232,8 @@ impl RedisClientInner {
     let id = Arc::new(format!("fred-{}", utils::random_string(10)));
     let resolver = DefaultResolver::new(&id);
     let (command_tx, command_rx) = unbounded_channel();
+    let version = config.version.clone();
+    let perf_config = InternalPerfConfig::from(&config);
 
     Arc::new(RedisClientInner {
       #[cfg(feature = "metrics")]
@@ -131,6 +245,8 @@ impl RedisClientInner {
       #[cfg(feature = "metrics")]
       res_size_stats: Arc::new(RwLock::new(MovingStats::default())),
 
+      resp_version: Arc::new(ArcSwap::from(Arc::new(version))),
+      perf_config: Arc::new(perf_config),
       config: RwLock::new(config),
       policy: RwLock::new(None),
       state: RwLock::new(ClientState::Disconnected),
@@ -138,6 +254,7 @@ impl RedisClientInner {
       message_tx: RwLock::new(VecDeque::new()),
       keyspace_tx: RwLock::new(VecDeque::new()),
       reconnect_tx: RwLock::new(VecDeque::new()),
+      cluster_change_tx: RwLock::new(VecDeque::new()),
       connect_tx: RwLock::new(VecDeque::new()),
       reconnect_sleep_jh: RwLock::new(None),
       cmd_buffer_len: Arc::new(AtomicUsize::new(0)),
@@ -155,7 +272,7 @@ impl RedisClientInner {
   }
 
   pub fn is_pipelined(&self) -> bool {
-    self.config.read().pipeline
+    self.perf_config.pipeline()
   }
 
   pub fn log_client_name_fn<F>(&self, level: log::Level, func: F)
@@ -203,4 +320,17 @@ impl RedisClientInner {
     let mut guard = self.command_rx.write();
     *guard = Some(rx);
   }
+
+  pub fn is_resp3(&self) -> bool {
+    *self.resp_version.as_ref().load().as_ref() == RespVersion::RESP3
+  }
+
+  pub fn switch_protocol_versions(&self, version: RespVersion) {
+    self.resp_version.as_ref().store(Arc::new(version))
+  }
+
+  pub fn reset_protocol_version(&self) {
+    let version = self.config.read().version.clone();
+    self.resp_version.as_ref().store(Arc::new(version));
+  }
 }
