@@ -1,5 +1,7 @@
 pub mod codec;
 pub mod connection;
+#[cfg(feature = "network-logs")]
+pub mod debug;
 pub mod tls;
 pub mod types;
 pub mod utils;
