@@ -5,16 +5,17 @@ use crate::multiplexer::utils as multiplexer_utils;
 use crate::multiplexer::{sentinel, ConnectionIDs};
 use crate::protocol::types::{RedisCommand, RedisCommandKind};
 use crate::types::*;
+use bytes::Bytes;
+use bytes_utils::Str;
 use float_cmp::approx_eq;
 use futures::future::{select, Either};
 use futures::{pin_mut, Future};
 use parking_lot::RwLock;
 use rand::distributions::Alphanumeric;
 use rand::{self, Rng};
-use redis_protocol::resp2::types::Frame as ProtocolFrame;
+use redis_protocol::resp3::types::Frame as Resp3Frame;
 use std::collections::HashMap;
 use std::convert::TryInto;
-use std::hash::Hasher;
 use std::ops::DerefMut;
 use std::sync::atomic::{AtomicUsize, Ordering};
 use std::sync::Arc;
@@ -23,11 +24,7 @@ use std::{f64, mem};
 use tokio::sync::oneshot::{channel as oneshot_channel, Receiver as OneshotReceiver};
 use tokio::sync::RwLock as AsyncRwLock;
 use tokio::time::sleep;
-
-#[cfg(feature = "index-map")]
-use indexmap::map::IndexMap;
-#[cfg(feature = "index-map")]
-use std::hash::Hash;
+use url::Url;
 
 #[cfg(any(feature = "full-tracing", feature = "partial-tracing"))]
 use crate::protocol::utils as protocol_utils;
@@ -35,9 +32,32 @@ use crate::protocol::utils as protocol_utils;
 use crate::trace;
 #[cfg(any(feature = "full-tracing", feature = "partial-tracing"))]
 use futures::TryFutureExt;
+#[cfg(feature = "serde-json")]
+use serde_json::Value;
 #[cfg(any(feature = "full-tracing", feature = "partial-tracing"))]
 use tracing_futures::Instrument;
 
+const REDIS_TLS_SCHEME: &'static str = "rediss";
+const REDIS_CLUSTER_SCHEME_SUFFIX: &'static str = "-cluster";
+const REDIS_SENTINEL_SCHEME_SUFFIX: &'static str = "-sentinel";
+const SENTINEL_NAME_QUERY: &'static str = "sentinelServiceName";
+const CLUSTER_NODE_QUERY: &'static str = "node";
+#[cfg(feature = "sentinel-auth")]
+const SENTINEL_USERNAME_QUERY: &'static str = "sentinelUsername";
+#[cfg(feature = "sentinel-auth")]
+const SENTINEL_PASSWORD_QUERY: &'static str = "sentinelPassword";
+
+/// Create a `Str` from a static str slice without copying.
+pub fn static_str(s: &'static str) -> Str {
+  // it's already parsed as a string
+  unsafe { Str::from_inner_unchecked(Bytes::from_static(s.as_bytes())) }
+}
+
+/// Create a `Bytes` from static bytes without copying.
+pub fn static_bytes(b: &'static [u8]) -> Bytes {
+  Bytes::from_static(b)
+}
+
 pub fn is_clustered(config: &RwLock<RedisConfig>) -> bool {
   config.read().server.is_clustered()
 }
@@ -80,9 +100,9 @@ pub fn redis_string_to_f64(s: &str) -> Result<f64, RedisError> {
 /// Convert an `f64` to a redis string, supporting "+inf" and "-inf".
 pub fn f64_to_redis_string(d: f64) -> Result<RedisValue, RedisError> {
   if d.is_infinite() && d.is_sign_negative() {
-    Ok("-inf".into())
+    Ok(RedisValue::from_static_str("-inf"))
   } else if d.is_infinite() {
-    Ok("+inf".into())
+    Ok(RedisValue::from_static_str("+inf"))
   } else if d.is_nan() {
     Err(RedisError::new(
       RedisErrorKind::InvalidArgument,
@@ -115,7 +135,7 @@ pub fn incr_with_max(curr: u32, max: u32) -> Option<u32> {
   if max != 0 && curr >= max {
     None
   } else {
-    Some(curr + 1)
+    Some(curr.saturating_add(1))
   }
 }
 
@@ -154,13 +174,13 @@ pub fn pattern_pubsub_counts(result: Vec<RedisValue>) -> Result<Vec<usize>, Redi
           } else {
             *i as usize
           }
-        }
+        },
         _ => {
           return Err(RedisError::new(
             RedisErrorKind::Unknown,
             "Invalid pattern pubsub response.",
           ))
-        }
+        },
       });
 
       idx += 3;
@@ -194,6 +214,14 @@ pub fn check_and_set_client_state(
   }
 }
 
+/// Check and set the inner locked value by updating `locked` with `new_value`, returning the old value.
+pub fn check_and_set_bool(locked: &RwLock<bool>, new_value: bool) -> bool {
+  let mut guard = locked.write();
+  let old_value = *guard;
+  *guard = new_value;
+  old_value
+}
+
 pub fn read_centralized_server(inner: &Arc<RedisClientInner>) -> Option<Arc<String>> {
   match inner.config.read().server {
     ServerConfig::Centralized { ref host, ref port } => Some(Arc::new(format!("{}:{}", host, port))),
@@ -202,43 +230,6 @@ pub fn read_centralized_server(inner: &Arc<RedisClientInner>) -> Option<Arc<Stri
   }
 }
 
-#[cfg(feature = "index-map")]
-pub fn new_map(capacity: usize) -> IndexMap<String, RedisValue> {
-  if capacity == 0 {
-    IndexMap::new()
-  } else {
-    IndexMap::with_capacity(capacity)
-  }
-}
-
-#[cfg(not(feature = "index-map"))]
-pub fn new_map(capacity: usize) -> HashMap<String, RedisValue> {
-  if capacity == 0 {
-    HashMap::new()
-  } else {
-    HashMap::with_capacity(capacity)
-  }
-}
-
-#[cfg(feature = "index-map")]
-pub fn hash_map<H>(data: &IndexMap<String, RedisValue>, state: &mut H)
-where
-  H: Hasher,
-{
-  for (key, value) in data.iter() {
-    key.hash(state);
-    value.hash(state);
-  }
-}
-
-#[cfg(not(feature = "index-map"))]
-pub fn hash_map<H>(_data: &HashMap<String, RedisValue>, _state: &mut H)
-where
-  H: Hasher,
-{
-  panic!("Cannot use HashMap as hash key.");
-}
-
 pub fn decr_atomic(size: &Arc<AtomicUsize>) -> usize {
   size.fetch_sub(1, Ordering::AcqRel).saturating_sub(1)
 }
@@ -409,7 +400,7 @@ pub fn send_command(inner: &Arc<RedisClientInner>, command: RedisCommand) -> Res
     decr_atomic(&inner.cmd_buffer_len);
     if let Some(tx) = e.0.tx.take() {
       if let Err(_) = tx.send(Err(RedisError::new(RedisErrorKind::Unknown, "Failed to send command."))) {
-        _error!(inner, "Failed to send command {:?}.", e.0.extract_key());
+        _error!(inner, "Failed to send command to multiplexer {:?}.", e.0.kind);
       }
     }
   }
@@ -437,10 +428,10 @@ where
 }
 
 async fn wait_for_response(
-  rx: OneshotReceiver<Result<ProtocolFrame, RedisError>>,
-) -> Result<ProtocolFrame, RedisError> {
-  let sleep_duration = globals().default_command_timeout();
-  apply_timeout(rx, sleep_duration as u64).await?
+  rx: OneshotReceiver<Result<Resp3Frame, RedisError>>,
+  timeout: u64,
+) -> Result<Resp3Frame, RedisError> {
+  apply_timeout(rx, timeout).await?
 }
 
 fn has_blocking_error_policy(inner: &Arc<RedisClientInner>) -> bool {
@@ -470,10 +461,10 @@ pub async fn interrupt_blocked_connection(
         RedisErrorKind::Unknown,
         "Failed to find blocked connection ID.",
       ))
-    }
+    },
   };
 
-  backchannel_request_response(inner, move || {
+  backchannel_request_response(inner, true, move || {
     Ok((
       RedisCommandKind::ClientUnblock,
       vec![connection_id.into(), flag.to_str().into()],
@@ -506,7 +497,7 @@ async fn check_blocking_policy(inner: &Arc<RedisClientInner>, command: &RedisCom
   Ok(())
 }
 
-pub async fn basic_request_response<F>(inner: &Arc<RedisClientInner>, func: F) -> Result<ProtocolFrame, RedisError>
+pub async fn basic_request_response<F>(inner: &Arc<RedisClientInner>, func: F) -> Result<Resp3Frame, RedisError>
 where
   F: FnOnce() -> Result<(RedisCommandKind, Vec<RedisValue>), RedisError>,
 {
@@ -518,11 +509,11 @@ where
   let _ = disallow_nested_values(&command)?;
   let _ = send_command(&inner, command)?;
 
-  wait_for_response(rx).await
+  wait_for_response(rx, inner.perf_config.default_command_timeout() as u64).await
 }
 
 #[cfg(any(feature = "full-tracing", feature = "partial-tracing"))]
-pub async fn request_response<F>(inner: &Arc<RedisClientInner>, func: F) -> Result<ProtocolFrame, RedisError>
+pub async fn request_response<F>(inner: &Arc<RedisClientInner>, func: F) -> Result<Resp3Frame, RedisError>
 where
   F: FnOnce() -> Result<(RedisCommandKind, Vec<RedisValue>), RedisError>,
 {
@@ -547,9 +538,6 @@ where
     let _ = disallow_nested_values(&command)?;
     (command, rx, req_size)
   };
-  if let Some(key) = command.extract_key() {
-    cmd_span.record("key", &&*key);
-  }
   cmd_span.record("cmd", &command.kind.to_str_debug());
   cmd_span.record("req_size", &req_size);
 
@@ -559,7 +547,7 @@ where
 
   let _ = check_blocking_policy(inner, &command).await?;
   let _ = send_command(&inner, command)?;
-  wait_for_response(rx)
+  wait_for_response(rx, inner.perf_config.default_command_timeout() as u64)
     .and_then(|frame| async move {
       trace::record_response_size(&end_cmd_span, &frame);
       Ok::<_, RedisError>(frame)
@@ -569,7 +557,7 @@ where
 }
 
 #[cfg(not(any(feature = "full-tracing", feature = "partial-tracing")))]
-pub async fn request_response<F>(inner: &Arc<RedisClientInner>, func: F) -> Result<ProtocolFrame, RedisError>
+pub async fn request_response<F>(inner: &Arc<RedisClientInner>, func: F) -> Result<Resp3Frame, RedisError>
 where
   F: FnOnce() -> Result<(RedisCommandKind, Vec<RedisValue>), RedisError>,
 {
@@ -592,7 +580,7 @@ fn find_backchannel_server(inner: &Arc<RedisClientInner>, command: &RedisCommand
           RedisErrorKind::Sentinel,
           "Failed to read sentinel primary server",
         ))
-    }
+    },
     ServerConfig::Centralized { ref host, ref port } => Ok(Arc::new(format!("{}:{}", host, port))),
     ServerConfig::Clustered { .. } => {
       if let Some(key) = command.extract_key() {
@@ -606,14 +594,14 @@ fn find_backchannel_server(inner: &Arc<RedisClientInner>, command: &RedisCommand
                 RedisErrorKind::Cluster,
                 "Failed to find cluster node at hash slot.",
               ))
-            }
+            },
           },
           None => {
             return Err(RedisError::new(
               RedisErrorKind::Cluster,
               "Failed to find cluster state.",
             ))
-          }
+          },
         };
 
         Ok(server)
@@ -627,26 +615,27 @@ fn find_backchannel_server(inner: &Arc<RedisClientInner>, command: &RedisCommand
                 RedisErrorKind::Cluster,
                 "Failed to find read random cluster node.",
               ))
-            }
+            },
           },
           None => {
             return Err(RedisError::new(
               RedisErrorKind::Cluster,
               "Failed to find cluster state.",
             ))
-          }
+          },
         };
 
         Ok(server)
       }
-    }
+    },
   }
 }
 
 pub async fn backchannel_request_response<F>(
   inner: &Arc<RedisClientInner>,
+  use_blocked: bool,
   func: F,
-) -> Result<ProtocolFrame, RedisError>
+) -> Result<Resp3Frame, RedisError>
 where
   F: FnOnce() -> Result<(RedisCommandKind, Vec<RedisValue>), RedisError>,
 {
@@ -669,7 +658,7 @@ where
       .backchannel
       .write()
       .await
-      .request_response(inner, blocked_server, command)
+      .request_response(inner, blocked_server, command, use_blocked)
       .await
   } else {
     // otherwise no connections are blocked
@@ -685,7 +674,7 @@ where
       .backchannel
       .write()
       .await
-      .request_response(inner, &server, command)
+      .request_response(inner, &server, command, use_blocked)
       .await
   }
 }
@@ -700,7 +689,7 @@ pub async fn read_connection_ids(inner: &Arc<RedisClientInner>) -> Option<HashMa
     .and_then(|connection_ids| match connection_ids {
       ConnectionIDs::Clustered(ref connection_ids) => {
         Some(connection_ids.read().iter().map(|(k, v)| (k.clone(), *v)).collect())
-      }
+      },
       ConnectionIDs::Centralized(connection_id) => {
         if let Some(id) = connection_id.read().as_ref() {
           let mut out = HashMap::with_capacity(1);
@@ -714,11 +703,10 @@ pub async fn read_connection_ids(inner: &Arc<RedisClientInner>) -> Option<HashMa
         } else {
           None
         }
-      }
+      },
     })
 }
 
-// TODO clean this up in the next major version
 pub fn read_sentinel_host(inner: &Arc<RedisClientInner>) -> Result<(Vec<(String, u16)>, String), RedisError> {
   match inner.config.read().server {
     #[cfg(not(feature = "sentinel-auth"))]
@@ -754,7 +742,7 @@ pub async fn update_sentinel_nodes(inner: &Arc<RedisClientInner>) -> Result<(),
           e
         );
         continue;
-      }
+      },
     };
     if let Err(e) = sentinel::update_sentinel_nodes(inner, transport, &service_name).await {
       _warn!(
@@ -852,3 +840,291 @@ where
 
   Ok(out)
 }
+
+pub fn add_jitter(delay: u64, jitter: u32) -> u64 {
+  delay.saturating_add(rand::thread_rng().gen_range(0..jitter as u64))
+}
+
+pub fn into_redis_map<I, K, V>(mut iter: I) -> Result<HashMap<RedisKey, RedisValue>, RedisError>
+where
+  I: Iterator<Item = (K, V)>,
+  K: TryInto<RedisKey>,
+  K::Error: Into<RedisError>,
+  V: TryInto<RedisValue>,
+  V::Error: Into<RedisError>,
+{
+  let (lower, upper) = iter.size_hint();
+  let capacity = if let Some(upper) = upper { upper } else { lower };
+  let mut out = HashMap::with_capacity(capacity);
+
+  while let Some((key, value)) = iter.next() {
+    out.insert(to!(key)?, to!(value)?);
+  }
+  Ok(out)
+}
+
+#[cfg(feature = "serde-json")]
+pub fn parse_nested_json(s: &str) -> Option<Value> {
+  let trimmed = s.trim();
+  let is_maybe_json =
+    (trimmed.starts_with("{") && trimmed.ends_with("}")) || (trimmed.starts_with("[") && trimmed.ends_with("]"));
+
+  if is_maybe_json {
+    serde_json::from_str(s).ok()
+  } else {
+    None
+  }
+}
+
+pub fn flatten_nested_array_values(value: RedisValue, depth: usize) -> RedisValue {
+  if depth == 0 {
+    return value;
+  }
+
+  match value {
+    RedisValue::Array(values) => {
+      let inner_size = values.iter().fold(0, |s, v| s + v.array_len().unwrap_or(1));
+      let mut out = Vec::with_capacity(inner_size);
+
+      for value in values.into_iter() {
+        match value {
+          RedisValue::Array(inner) => {
+            for value in inner.into_iter() {
+              out.push(flatten_nested_array_values(value, depth - 1));
+            }
+          },
+          _ => out.push(value),
+        }
+      }
+      RedisValue::Array(out)
+    },
+    RedisValue::Map(values) => {
+      let mut out = HashMap::with_capacity(values.len());
+
+      for (key, value) in values.inner().into_iter() {
+        let value = if value.is_array() {
+          flatten_nested_array_values(value, depth - 1)
+        } else {
+          value
+        };
+
+        out.insert(key, value);
+      }
+      RedisValue::Map(RedisMap { inner: out })
+    },
+    _ => value,
+  }
+}
+
+pub fn is_maybe_array_map(arr: &Vec<RedisValue>) -> bool {
+  if arr.len() > 0 && arr.len() % 2 == 0 {
+    arr.chunks(2).fold(true, |b, chunk| b && !chunk[0].is_aggregate_type())
+  } else {
+    false
+  }
+}
+
+#[cfg(feature = "enable-tls")]
+pub fn check_tls_features() {}
+
+#[cfg(not(feature = "enable-tls"))]
+pub fn check_tls_features() {
+  warn!("TLS features are not enabled, but a TLS feature may have been used.");
+}
+
+#[cfg(feature = "enable-tls")]
+pub fn tls_config_from_url(tls: bool) -> Option<TlsConfig> {
+  if tls {
+    Some(TlsConfig::default())
+  } else {
+    None
+  }
+}
+
+pub fn url_uses_tls(url: &Url) -> bool {
+  url.scheme().starts_with(REDIS_TLS_SCHEME)
+}
+
+pub fn url_is_clustered(url: &Url) -> bool {
+  url.scheme().ends_with(REDIS_CLUSTER_SCHEME_SUFFIX)
+}
+
+pub fn url_is_sentinel(url: &Url) -> bool {
+  url.scheme().ends_with(REDIS_SENTINEL_SCHEME_SUFFIX)
+}
+
+pub fn parse_url(url: &str, default_port: Option<u16>) -> Result<(Url, String, u16, bool), RedisError> {
+  let url = Url::parse(url)?;
+  let host = if let Some(host) = url.host_str() {
+    host.to_owned()
+  } else {
+    return Err(RedisError::new(RedisErrorKind::Config, "Invalid or missing host."));
+  };
+  let port = if let Some(port) = url.port().or(default_port) {
+    port
+  } else {
+    return Err(RedisError::new(RedisErrorKind::Config, "Invalid or missing port."));
+  };
+
+  let tls = url_uses_tls(&url);
+  if tls {
+    check_tls_features();
+  }
+
+  Ok((url, host, port, tls))
+}
+
+pub fn parse_url_db(url: &Url) -> Result<Option<u8>, RedisError> {
+  let parts: Vec<&str> = if let Some(parts) = url.path_segments() {
+    parts.collect()
+  } else {
+    return Ok(None);
+  };
+
+  if parts.len() > 1 {
+    return Err(RedisError::new(RedisErrorKind::Config, "Invalid database path."));
+  } else if parts.is_empty() {
+    return Ok(None);
+  }
+  // gracefully handle empty paths with a / prefix
+  if parts[0].trim() == "" {
+    return Ok(None);
+  }
+
+  Ok(Some(parts[0].parse()?))
+}
+
+pub fn parse_url_credentials(url: &Url) -> (Option<String>, Option<String>) {
+  let username = if url.username().is_empty() {
+    None
+  } else {
+    Some(url.username().to_owned())
+  };
+  let password = url.password().map(|s| s.to_owned());
+
+  (username, password)
+}
+
+pub fn parse_url_other_nodes(url: &Url) -> Result<Vec<(String, u16)>, RedisError> {
+  let mut out = Vec::new();
+
+  for (key, value) in url.query_pairs().into_iter() {
+    if key == CLUSTER_NODE_QUERY {
+      let parts: Vec<&str> = value.split(":").collect();
+      if parts.len() != 2 {
+        return Err(RedisError::new(
+          RedisErrorKind::Config,
+          format!("Invalid host:port for cluster node: {}", value),
+        ));
+      }
+
+      let host = parts[0].to_owned();
+      let port = parts[1].parse::<u16>()?;
+      out.push((host, port));
+    }
+  }
+
+  Ok(out)
+}
+
+pub fn parse_url_sentinel_service_name(url: &Url) -> Result<String, RedisError> {
+  for (key, value) in url.query_pairs().into_iter() {
+    if key == SENTINEL_NAME_QUERY {
+      return Ok(value.to_string());
+    }
+  }
+
+  Err(RedisError::new(
+    RedisErrorKind::Config,
+    "Invalid or missing sentinel service name query parameter.",
+  ))
+}
+
+#[cfg(feature = "sentinel-auth")]
+pub fn parse_url_sentinel_username(url: &Url) -> Option<String> {
+  url.query_pairs().find_map(|(key, value)| {
+    if key == SENTINEL_USERNAME_QUERY {
+      Some(value.to_string())
+    } else {
+      None
+    }
+  })
+}
+
+#[cfg(feature = "sentinel-auth")]
+pub fn parse_url_sentinel_password(url: &Url) -> Option<String> {
+  url.query_pairs().find_map(|(key, value)| {
+    if key == SENTINEL_PASSWORD_QUERY {
+      Some(value.to_string())
+    } else {
+      None
+    }
+  })
+}
+
+#[cfg(test)]
+mod tests {
+  use super::*;
+  use crate::error::RedisError;
+  use crate::types::RedisValue;
+  use std::convert::TryInto;
+  use std::fmt::Debug;
+
+  fn m<V>(v: V) -> RedisValue
+  where
+    V: TryInto<RedisValue> + Debug,
+    V::Error: Into<RedisError> + Debug,
+  {
+    v.try_into().unwrap()
+  }
+
+  fn a(v: Vec<RedisValue>) -> RedisValue {
+    RedisValue::Array(v)
+  }
+
+  #[test]
+  fn should_flatten_xread_example() {
+    // 127.0.0.1:6379> xread count 2 streams foo bar 1643479648480-0 1643479834990-0
+    // 1) 1) "foo"
+    //    2) 1) 1) "1643479650336-0"
+    //          2) 1) "count"
+    //             2) "3"
+    // 2) 1) "bar"
+    //    2) 1) 1) "1643479837746-0"
+    //          2) 1) "count"
+    //             2) "5"
+    //       2) 1) "1643479925582-0"
+    //          2) 1) "count"
+    //             2) "6"
+    let actual: RedisValue = vec![
+      a(vec![
+        m("foo"),
+        a(vec![a(vec![m("1643479650336-0"), a(vec![m("count"), m(3)])])]),
+      ]),
+      a(vec![
+        m("bar"),
+        a(vec![
+          a(vec![m("1643479837746-0"), a(vec![m("count"), m(5)])]),
+          a(vec![m("1643479925582-0"), a(vec![m("count"), m(6)])]),
+        ]),
+      ]),
+    ]
+    .into_iter()
+    .collect();
+
+    // flatten the top level nested array into something that can be cast to a map
+    let expected: RedisValue = vec![
+      m("foo"),
+      a(vec![a(vec![m("1643479650336-0"), a(vec![m("count"), m(3)])])]),
+      m("bar"),
+      a(vec![
+        a(vec![m("1643479837746-0"), a(vec![m("count"), m(5)])]),
+        a(vec![m("1643479925582-0"), a(vec![m("count"), m(6)])]),
+      ]),
+    ]
+    .into_iter()
+    .collect();
+
+    assert_eq!(flatten_nested_array_values(actual, 1), expected);
+  }
+}
