@@ -1,7 +1,7 @@
-use fred::client::RedisClient;
+use fred::clients::RedisClient;
 use fred::error::RedisError;
-use fred::types::{RedisConfig, RedisMap, RedisValue};
-
+use fred::interfaces::*;
+use fred::types::{RedisConfig, RedisValue};
 use std::collections::{HashMap, HashSet};
 
 fn assert_contains<'a, T: Eq + PartialEq>(values: Vec<T>, item: &'a T) {
@@ -36,9 +36,9 @@ fn assert_diff_len(values: Vec<&'static str>, value: RedisValue, len: usize) {
 pub async fn should_hset_and_hget(client: RedisClient, _: RedisConfig) -> Result<(), RedisError> {
   check_null!(client, "foo");
 
-  let result: i64 = client.hset("foo", ("a", 1.into())).await?;
+  let result: i64 = client.hset("foo", ("a", 1)).await?;
   assert_eq!(result, 1);
-  let result: i64 = client.hset("foo", vec![("b", 2.into()), ("c", 3.into())]).await?;
+  let result: i64 = client.hset("foo", vec![("b", 2), ("c", 3)]).await?;
   assert_eq!(result, 2);
 
   let a: i64 = client.hget("foo", "a").await?;
@@ -54,9 +54,7 @@ pub async fn should_hset_and_hget(client: RedisClient, _: RedisConfig) -> Result
 pub async fn should_hset_and_hdel(client: RedisClient, _: RedisConfig) -> Result<(), RedisError> {
   check_null!(client, "foo");
 
-  let result: i64 = client
-    .hset("foo", vec![("a", 1.into()), ("b", 2.into()), ("c", 3.into())])
-    .await?;
+  let result: i64 = client.hset("foo", vec![("a", 1.into()), ("b", 2), ("c", 3)]).await?;
   assert_eq!(result, 3);
   let result: i64 = client.hdel("foo", vec!["a", "b"]).await?;
   assert_eq!(result, 2);
@@ -71,7 +69,7 @@ pub async fn should_hset_and_hdel(client: RedisClient, _: RedisConfig) -> Result
 pub async fn should_hexists(client: RedisClient, _: RedisConfig) -> Result<(), RedisError> {
   check_null!(client, "foo");
 
-  let _: () = client.hset("foo", ("a", 1.into())).await?;
+  let _: () = client.hset("foo", ("a", 1)).await?;
   let a: bool = client.hexists("foo", "a").await?;
   assert!(a);
   let b: bool = client.hexists("foo", "b").await?;
@@ -83,9 +81,7 @@ pub async fn should_hexists(client: RedisClient, _: RedisConfig) -> Result<(), R
 pub async fn should_hgetall(client: RedisClient, _: RedisConfig) -> Result<(), RedisError> {
   check_null!(client, "foo");
 
-  let _: () = client
-    .hset("foo", vec![("a", 1.into()), ("b", 2.into()), ("c", 3.into())])
-    .await?;
+  let _: () = client.hset("foo", vec![("a", 1), ("b", 2), ("c", 3)]).await?;
   let values: HashMap<String, i64> = client.hgetall("foo").await?;
 
   assert_eq!(values.len(), 3);
@@ -123,9 +119,7 @@ pub async fn should_hincryby_float(client: RedisClient, _: RedisConfig) -> Resul
 pub async fn should_get_keys(client: RedisClient, _: RedisConfig) -> Result<(), RedisError> {
   check_null!(client, "foo");
 
-  let _: () = client
-    .hset("foo", vec![("a", 1.into()), ("b", 2.into()), ("c", 3.into())])
-    .await?;
+  let _: () = client.hset("foo", vec![("a", 1), ("b", 2), ("c", 3)]).await?;
 
   let keys = client.hkeys("foo").await?;
   assert_diff_len(vec!["a", "b", "c"], keys, 0);
@@ -136,9 +130,7 @@ pub async fn should_get_keys(client: RedisClient, _: RedisConfig) -> Result<(),
 pub async fn should_hmset(client: RedisClient, _: RedisConfig) -> Result<(), RedisError> {
   check_null!(client, "foo");
 
-  let _: () = client
-    .hmset("foo", vec![("a", 1.into()), ("b", 2.into()), ("c", 3.into())])
-    .await?;
+  let _: () = client.hmset("foo", vec![("a", 1), ("b", 2), ("c", 3)]).await?;
 
   let a: i64 = client.hget("foo", "a").await?;
   assert_eq!(a, 1);
@@ -153,9 +145,7 @@ pub async fn should_hmset(client: RedisClient, _: RedisConfig) -> Result<(), Red
 pub async fn should_hmget(client: RedisClient, _: RedisConfig) -> Result<(), RedisError> {
   check_null!(client, "foo");
 
-  let _: () = client
-    .hmset("foo", vec![("a", 1.into()), ("b", 2.into()), ("c", 3.into())])
-    .await?;
+  let _: () = client.hmset("foo", vec![("a", 1), ("b", 2), ("c", 3)]).await?;
 
   let result: Vec<i64> = client.hmget("foo", vec!["a", "b"]).await?;
   assert_eq!(result, vec![1, 2]);
@@ -166,7 +156,7 @@ pub async fn should_hmget(client: RedisClient, _: RedisConfig) -> Result<(), Red
 pub async fn should_hsetnx(client: RedisClient, _: RedisConfig) -> Result<(), RedisError> {
   check_null!(client, "foo");
 
-  let _: () = client.hset("foo", ("a", 1.into())).await?;
+  let _: () = client.hset("foo", ("a", 1)).await?;
   let result: bool = client.hsetnx("foo", "a", 2).await?;
   assert_eq!(result, false);
   let result: i64 = client.hget("foo", "a").await?;
@@ -182,9 +172,7 @@ pub async fn should_hsetnx(client: RedisClient, _: RedisConfig) -> Result<(), Re
 pub async fn should_get_random_field(client: RedisClient, _: RedisConfig) -> Result<(), RedisError> {
   check_null!(client, "foo");
 
-  let _: () = client
-    .hmset("foo", vec![("a", 1.into()), ("b", 2.into()), ("c", 3.into())])
-    .await?;
+  let _: () = client.hmset("foo", vec![("a", 1), ("b", 2), ("c", 3)]).await?;
 
   let field: String = client.hrandfield("foo", None).await?;
   assert_contains(vec!["a", "b", "c"], &field.as_str());
@@ -195,13 +183,13 @@ pub async fn should_get_random_field(client: RedisClient, _: RedisConfig) -> Res
   let actual: HashMap<String, i64> = client.hrandfield("foo", Some((2, true))).await?;
   assert_eq!(actual.len(), 2);
 
-  let mut expected = RedisMap::new();
-  expected.insert("a".into(), "1".into());
-  expected.insert("b".into(), "2".into());
-  expected.insert("c".into(), "3".into());
+  let mut expected: HashMap<String, i64> = HashMap::new();
+  expected.insert("a".into(), 1);
+  expected.insert("b".into(), 2);
+  expected.insert("c".into(), 3);
 
   for (key, value) in actual.into_iter() {
-    let expected_val: i64 = expected.get(&key).unwrap().clone().convert()?;
+    let expected_val: i64 = *expected.get(&key).unwrap();
     assert_eq!(value, expected_val);
   }
 
@@ -212,7 +200,7 @@ pub async fn should_get_strlen(client: RedisClient, _: RedisConfig) -> Result<()
   check_null!(client, "foo");
 
   let expected = "abcdefhijklmnopqrstuvwxyz";
-  let _: () = client.hset("foo", ("a", expected.clone().into())).await?;
+  let _: () = client.hset("foo", ("a", expected)).await?;
 
   let len: usize = client.hstrlen("foo", "a").await?;
   assert_eq!(len, expected.len());
@@ -223,7 +211,7 @@ pub async fn should_get_strlen(client: RedisClient, _: RedisConfig) -> Result<()
 pub async fn should_get_values(client: RedisClient, _: RedisConfig) -> Result<(), RedisError> {
   check_null!(client, "foo");
 
-  let _: () = client.hmset("foo", vec![("a", "1".into()), ("b", "2".into())]).await?;
+  let _: () = client.hmset("foo", vec![("a", "1"), ("b", "2")]).await?;
 
   let values: RedisValue = client.hvals("foo").await?;
   assert_diff_len(vec!["1", "2"], values, 0);
