@@ -1,6 +1,11 @@
+#![allow(unused_imports)]
+
 #[macro_use]
 extern crate log;
 extern crate pretty_env_logger;
 
+#[macro_use]
+extern crate maplit;
+
 mod chaos_monkey;
 mod integration;
